
<div class="abc" style="float: left; margin-left: 15%; margin-right:15%; width: 70%" >
<h2>Dynaseq Analysis   <i class="icon-help-circled" onClick="toggle1(this.parentNode.parentNode)" style="cursor:pointer;"></i></h2>
<div class="box">
    <div class="description" style="display: none;">
        <p><b>Substitution rate</b> : Overall nucleotide substitution rates in a sample.</p>
        <p><b>Turn-over rate</b> : Labeled transcripts (UMI) fraction per cell or per gene.</p>
    </div>


<h3>Sample substitution rate</h3>
{{ step_summary.substitution|safe }}


